/*
   Sketch / Application to read data on HP 2748 paper tape reader, and
   send it to a TCP client.

   PINOUTS

    A3     Block
    Pin     Pin     Name       Color         DB25 (Parallel)      Arduino Pin

   XA3 AA    AA      /READ(cmd) GRN             6                   A4 / PC4
   XA3 X     FF      /HOLE      RED             8                   A1 / PC1
   XA3 P     M       /EOT                      10                   A2 / PC2
   XA321     S       /READY                    12                   A3 / PC3
   XA3 1     B       /BIT1(LSB) BLU/BLK        13                   D2
   XA3 3     F       /BIT2      BRN/BLK        14                   D3
   XA3 5     L       /BIT3      ORG/BLK        15                   D5
   XA3 7     R       /BIT4      YEL/BLK        16                   D6
   XA3 9     V       /BIT5      RED/BLK        17                   D7
   XA3 12    Z       /BIT6      WHT/BLK        18                   D8 / PB0
   XA3 S     d       /BIT7      GRN/BLK        19                   D9 / PB2
   XA3 V     j       /BIT8(MSB) WHT/RED        20                   A0 / PC0
   Chassis   HH      GND        BLK             7                   GND

   Arduino Register Assignments

   PORTD  Bits 543.21..   (. == used for other purposes / unused)
   PORTB  Bits ......76   (. == used for other purposes / unused)
   PORTC  Bits ...RYEH8   (R == Read, H == Hole, E == EOT, Y = READY)

   Copyright (C) 2023 Jay R. Jaeger
   This program may be freely shared so long as credit is given to the author

*/

#include <SPI.h>
#include <EthernetV2_0.h>

#define DEBUG 0
#define GETDATA ((PINC & 0x01) << 7) | ((PINB & 0x03) << 5) | \
  ((PIND & 0xe0) >> 3) | ((PIND & 0x0c) >> 2)

#define HOLEMASK 2
#define READMASK 0x10
#define PTRBLOCKSZ 128
#define EOTMASK  4
#define RDYMASK  8

#define W5200_CS 10
#define SDCARD_CS 4

unsigned char buf[PTRBLOCKSZ];

//  Mac Address 0xDE 0xAD 'j' 'r' 'j' #Sequence Number
byte mac[] = {
  0xDE, 0xAD, 0x6A, 0x72, 0x6A, 0x01
};
IPAddress ip(192, 168, 42, 254);

EthernetServer server = EthernetServer(10001);

void setup() {

  //  Set up ports for appropriate directions - mostly IN

  //  PORTC is A5-A0: A0-A3 are IN, A4 is OUT

  DDRC = DDRC & B01110000;
  DDRC = DDRC | B00010000;

  //  PORTB is D13-D8: D8 and D9 are IN

  DDRB = DDRB & B11111100;

  //  PORTD is D7-D0, D2, D3, D5-D7 are IN

  DDRD = DDRD & B00010011;

  //  Set up serial console

  Serial.begin(9600);

  //  Disable the SD Card
  pinMode(SDCARD_CS, OUTPUT);
  digitalWrite(SDCARD_CS, HIGH);

  //  Start up Ethernet 
  Ethernet.begin(mac, ip);
  server.begin();
  Serial.print("HP2874 PTR is ready at ");
  Serial.println(Ethernet.localIP());
}

void sendBuffer(int count, EthernetClient client) {
  unsigned char header[3];
  // header[0] = 0xee;
  // header[1] = count & 0xff;
  // header[2] = (count >> 8) & 0xff;
  // client.write(header,sizeof header);
  client.write(buf,count);
}

void loop() {

  int i;
  unsigned char *p;
  int blkcnt = 0;

  //  Set READ to INACTIVE

  PORTC = PORTC | READMASK;

  EthernetClient client;

  //  Wait for TCP Connection

  Serial.println("Waiting for client connection.");

  while ((client = server.available()) == 0)
    ;

   Serial.println("TCP Connection made.");
  
  //  Now, go into a loop reading blocks.  Each Block is PTRBLOCKSZ bytes.
  //  Last block may be short.  Data sent is just the data, with
  //  no header or anything like that.HP
  
  p = buf;                                  // Point to start of buffer
  blkcnt = 0;                               // No bytes received so far

  while (1) {
 
    //  Wait for READY

    // Serial.println("Waiting for READY...");

    while((PINC & RDYMASK) != 0)
      ;
      
    // Activate read

    // Serial.println("Activating READ...");
   
    PORTC = PORTC & ~READMASK;               // Set /READ active

    while ((PINC & RDYMASK) == 0 &&         // WAIT while Ready AND
           (PINC & EOTMASK) != 0 &&         // NOT End of Tape
           (PINC & HOLEMASK) != 0)          // AND NOT a hole
      ;

    // If we hit end of tape, send any unsent data 
    
    if((PINC & EOTMASK) == 0) {
      if(blkcnt > 0) {
        sendBuffer(blkcnt,client);
      }
      Serial.println("End of Tape sensed");
      break;
    }

    //  If we went not ready for some reason, then just loop back

    if((PINC & RDYMASK) != 0) {
      continue;
    }

    //  If we get here, we should sense a hole...

    if((PINC & HOLEMASK)!= 0) {
      Serial.println("Hole then not hole too fast...");
      continue;
    }

    //  Read the data into the buffer...

    *p++ = ~(GETDATA);
    ++blkcnt;
   
    //  (If the buffer is full, drop the READ command, and)
    //  send the buffer on its way...  (READ will be asserted
    //  again near the top of the loop.)

    if(p - buf >= PTRBLOCKSZ) {
      // PORTC = PORTC | READMASK;              // Set /READ inactive
      // Serial.println("Sending Buffer...");
      sendBuffer(blkcnt,client);
      p = buf;
      // delay(100);
      blkcnt = 0;
    }

    //  Wait for the hole to go away (or end of tape)

    while((PINC & HOLEMASK) == 0 && (PINC & EOTMASK) != 0)
      ;

    //  Wait for a tiny bit longer unless we were at the end of a block
    //  This helps avoid doubling characters from stretched holes...
    
    if(blkcnt != 0) {
      for(i=0; i < 512; ++i)
        ;
    }


  }

  // Make sure everything gets sent before we disconnect.
 
  client.flush();

  //  All done...

  client.stop();

  Serial.println("Done. ");
  
}
