#!/usr/bin/perl

#
#	Verify an Altair/8080 Binary Tape
#
#	Format for a record:
#
#	0xFF, Addr High, Addr Low, RecordLength <RecordLength Bytes> Checksum
#

use Getopt::Std;

my $debug = 1;
my $self = "AltairCheckBin.pl";
my $word;

our($opt_d, $opt_o);;

sub usage();
sub dosum($$);

getopts('dvo:') || usage();
$debug = $opt_d;
$verbose = $opt_v;

if($opt_o) {
	$imageFile = $opt_o;
}

if($#ARGV < 0) {
	usage();
	exit(8);
}

open(INFILE,"<$ARGV[0]") || 
	die("Cannot open input file $ARGV[0]\n");
	
binmode(INFILE);

if($opt_o) {
	open(IMAGEFILE,">$imageFile") || 
		die("Cannot open output file $imageFile\n");
	binmode(IMAGEFILE);
}

print("$self: Processing file $ARGV[0]\n");


my $frame = 0;
my $doneflag = 0;
my $fileOffset = 0;
my $skipping = 0;
my $first = 1;
my $errorCount = 0;

$image[0] = -1;

while(1) {
	
	#
	#	Look for 0xFF record start, skipping any nul bytes
	#	

	while($frame != 0xff) {
		if(read(INFILE,$ch,1) != 1) {
			print("$self: EOF looking for next record.\n");
			$doneflag = 1;
			last;
		}
		$frame = ord($ch);

		if(!$first && !$skipping && $frame != 0xff) {
			printf("%s: Expected 0xFF at offset %5d (06%o), got %2x\n",
				$self,$fileOffset,$fileOffset,$frame);
			$skipping = 1;
			++$errorCount;
		}

		$recordOffset = $fileOffset;
		++$fileOffset;

	}

	if($doneflag) {
		last;
	}

	#	The flag byte isn't included in the checksum.

	$chksum = 0;


	#
	#	Found start of record, read address and count
	#

	$first = 0;
	if(read(INFILE,$ch,1) != 1) {
		print("$self: EOF reading high byte of address.\n");
		last;
	}
	$addrhigh = ord($ch);
	$chksum = dosum($chksum,$addrhigh);
	++$fileOffset;
	
	if(read(INFILE,$ch,1) != 1) {
		print("$self: EOF reading low byte of address.\n");
		last;
	}
	$addrlow = ord($ch);
	$chksum = dosum($chksum,$addrlow);
	++$fileOffset;
	
	if(read(INFILE,$ch,1) != 1) {
		print("$self: EOF reading byte count.\n");
		last;
	}
	$count = ord($ch);
	$chksum = dosum($chksum,$count);
	$imageAddress = ($addrhigh*256) + $addrlow;
	++$fileOffset;

	printf("%s: Offset %5d (0%06o) Addr %03o/%03o (%06o), Len %d\n",
		$self,$recordOffset,$recordOffset,$addrhigh,$addrlow,($addrhigh*256)+$addrlow,$count);

	#	Read the data bytes

	for($i = 0; $i < $count; ++$i) {
		if(read(INFILE,$ch,1) != 1) {
			print("$self: Error/EOF reading data byte in record.\n");
			last;
		}	
		$chksum = dosum($chksum, ord($ch));
		if($opt_o) {
			$image[$imageAddress] = ord($ch);
			++$imageAddress;
		}
		++$fileOffset;
	}	
	
	#	Read the checksum

	if(read(INFILE,$ch,1) != 1) {
		print("$self: EOF reading checksum.\n");
		last;
	}
	++$fileOffset;


	# printf("%s: Calculated: %03o, Got: %03o\n",$self,$chksum,ord($ch));
	$chksum = dosum($chksum,ord($ch));

	if($chksum != $recordSum) {
		printf("%s: Checksum mismatch: %03o\n",$self,$chksum);
		++$errorCount;
	}

	$frame = 0;
	$skipping = 0;
}

if($errorCount > 0) {
	print("$self: $errorCount ERRORS FOUND.\n");
}

close(INFILE);

if($opt_o && $#image > 0) {
	print "$self: saving image file $imageFile\n";
	for($i = 0; $i < $#image; ++$i) {
		print IMAGEFILE chr($image[$i]);
	}
	close(IMAGEFILE);
}

exit(0);

sub dosum($$) {
	my($sum, $added) = @_;
	# return ($sum + $added) % 255;
	return ($sum ^ $added);
}

sub usage() {
	print("Usage:  perl $self [-d] [-o imagefile] inputFile ");
}


